/* 
 * Naive XML Parser for RSS\RDF documents.
 * 2002-2003 HShivanan
 *
 * For licensing terms, see the file COPYING
 * For contact information, see the file AUTHORS
 *
 */


/* This expat-based parser is intentionally ultra-liberal and
 * doesn't perform any sort of validation beyond what is done
 * by expat for well-formed XML. 
 * This is because there are many different rdf\rss structures used
 * to define news feeds and not all of them have correctly defined
 * versioning metadata.
 *
 * We (try to) avoid problems by just looking at the elements we
 * have explicit interest in and ignoring the rest
 */


/*TODO: Use a table driven state-machine for parsing*/
#include <stdio.h>
#include <string.h>
#include "stuff.h"
#include "item.h"
#include "fn.h"
#include "expat/xmlparse.h"
#include "stream.h"

typedef enum {
	none = 0,
	title,
	description,
	link
}state_e;

typedef enum {
	no_context = 0,
	channel_context,
	item_context
}context_e;

typedef struct {
	state_e state;
	STREAM * stream;
	ITEM_LIST * item_list;
	ITEM * item;
	context_e context;
	int parse_error;
	char * channel;
}XML_STATE;

context_e get_context_from_str (const char *name)
{
	if (!strcasecmp(name,"item"))
		return item_context;
	if (!strcasecmp(name,"channel"))
		return channel_context;
	return no_context;
}

state_e get_state_from_str (const char *name)
{
	if (!strcasecmp(name,"description"))
		return description;
	if (!strcasecmp(name,"link"))
		return link;
	if (!strcasecmp(name,"title"))
		return title;
	return none;
}

void start_element (void * data, const char * name, const char ** atts)
{
	XML_STATE * xml_state = (XML_STATE *)data;
	context_e context = no_context;
	
	/*If we are entering a new context, just return */
	if ( (context = get_context_from_str (name)) != no_context)
	{
		xml_state->context = context;
		return;
	}
	
	if (xml_state->context == no_context)
		return;
	
	xml_state->state = get_state_from_str(name);

	xml_state->stream = stream_create (1024);
}

void char_handler (void * data, const char * name, int len)
{
	XML_STATE * xml_state = (XML_STATE *)data;

	if (xml_state->context == no_context)
		return;

	switch (xml_state->state)
	{
		case title: 
		case link:
		case description:
			stream_write (xml_state->stream, name, len);break;
		case none:
			   
	}
}

void end_element (void * data, const char * name)
{
	XML_STATE * xml_state = (XML_STATE *)data;
	context_e context = no_context;
	state_e state = none;

	context = get_context_from_str (name);
	
	if (context != no_context)
	{
		switch (context)
		{
			case channel_context:break;
			case item_context:
								 xml_state->item->channel = xml_state->channel;
								 add_item_item (&(xml_state->item_list),xml_state->item);
								 /*TODO: Memory leak here */
								 xml_state->item = (ITEM *)fn_malloc (sizeof(ITEM)); 
								 xml_state->item->channel = NULL;
								 xml_state->item->link = NULL;
								 xml_state->item->description = NULL;
								 xml_state->item->item = NULL;
								 if (xml_state->stream)
								 {
									 stream_free (xml_state->stream);
									 xml_state->stream = stream_create (1024);
								 }
								 break;
			case no_context:
								 ASSERT (1==0);
								 break;
		}
		return;
	}
	
	state = get_state_from_str (name);
	
	if (xml_state->state != state)
	{
		xml_state->parse_error = 1;
		return;
	}
	
	if (state!=none)
	{
		if (xml_state->context == no_context)
		{
			xml_state->parse_error = 1;
			return;
		}

		stream_add_char (xml_state->stream, '\0');
		switch (xml_state->state)
		{
			case title:
				if (xml_state->context == item_context)
					xml_state->item->item = stream_copy_buffer (xml_state->stream);
				else
					xml_state->channel = stream_copy_buffer (xml_state->stream);
				break;
			case description:
			if (xml_state->context == item_context)
					xml_state->item->description = stream_copy_buffer (xml_state->stream);
				break;
			case link:
				if (xml_state->context == item_context)
					xml_state->item->link = stream_copy_buffer (xml_state->stream);
				break;
			case none:
				ASSERT(1==0);
		}
		xml_state->state = none;
		
	}
	
	context = get_context_from_str (name);
}
void parse_newsfeed(ITEM_LIST **il,char *buffer,char *url,char *description)
{
	int result;
	XML_STATE xml_state;
	XML_Parser parser = XML_ParserCreate (NULL);
	xml_state.item = malloc(sizeof(ITEM));
	xml_state.item->description = NULL;
	xml_state.item->item = NULL;
	xml_state.item->channel = NULL;
	xml_state.item->link = NULL;
	xml_state.item_list  = *il;
	xml_state.parse_error = 0;
	xml_state.channel = NULL;
	xml_state.context = no_context;
	xml_state.state = none;
	XML_SetUserData(parser, &xml_state);
	XML_SetElementHandler (parser, start_element, end_element);
	XML_SetCharacterDataHandler (parser, char_handler);
	result = XML_Parse(parser, buffer, strlen(buffer),1);
	if (!result || xml_state.parse_error)
	{
#ifdef EXPAT_DEBUG
		error ("XML Parser:%s on line %d\n",
				XML_ErrorString(XML_GetErrorCode(parser)),
				XML_GetCurrentLineNumber(parser));
#else
		error("An Error occured while parsing %s:%d\n",
				url,
				XML_GetCurrentLineNumber(parser));
#endif
		return;
	}
	XML_ParserFree(parser);
}
